if [ -z ${CONDA_BUILD+x} ]; then
    source /opt/concourse/worker/volumes/live/8fc79583-b553-450a-5a86-0b6d58b92603/volume/pymongo_1573670002297/work/build_env_setup.sh
fi
python setup.py install --single-version-externally-managed --record=record.txt